// LCD_Driver.c Implementation of the HD44780 or simular LCD driver.
// The driver uses an 4 BIT DATA WORD. The DATA WORD is setup to use PORT C. To change it to another port
// Set DATA_PORT,TRIS_DATA_PORTC definition to new port in the header file to the port you want to use.

// The contol pins are setup to use PORT C. To Change the port to another port change the following control pin
// definitions in the header file.
// #define RS_COMMAND 		RC5 = 0
// #define RS_CHAR_DISPLAY 	RC5 = 1
// #define RW_WRITE		RC6 = 0
// #define RW_READ 		RC6 = 1
// #define CLOCK_L 		RC7 = 0
// #define CLOCK_H 		RC2 = 1

// #define RS_PIN_TO_OUTPUT  TRISC0 = 0
// #define RW_PIN_TO_OUTPUT  TRISC1 = 0
// #define CLK_PIN_TO_OUTPUT TRISC2 = 0

// There are 3 delay function in the library these need to be ammened for the clock being used. The fuctions are
// setup to give the delays with a 18.432MHz clock. The delay calls are based on the number of instruction cycles.
// 1 instrction cycle = 4/clock frequency.

// The delay fuction which need to be changed are
// ClockPulseH() 150us
// ClockPulseL() 150us
// Delay15ms()
// Delay5ms()
// Delay150us()

// To use the library first run the  InitDisplay(unsigned char display_lines,unsigned char font_size) function
// and set the number of display lines options are
// DISPLAY_LINES_1
// DISPLAY_LINES_2

// and the font type supported by the display. The options are
// FONT_5x8_DOT
// FONT_5x10_DOT

// Once you have initialized the display you can use any of the function avaliable in the library.
// Example to turn the display on and write a string to the display.
//
// #include <stdio.h>
//
// InitDisplay(DISLPAY_LINES_2,FONT_5x8_DOT);
		
//		while(BusyFlag());
//			DisplayOn();
	
//		sprintf((unsigned char *)t,"Desitronics");
//		number_of_bytes = sizeof((char *)t);

//		while(BusyFlag());
//			WriteDisplay(&t,number_of_bytes);

// Use the while(BusyFlag()) to check the display can accept a new command.

#include <xc.h>
#include <pic16.h>
#include <stdint.h>
#include "lcddriver.h"

void InitDisplay(unsigned char display_lines,unsigned char font_size)
{
	// Set Data word port to output
	TRIS_DATA_PORTC = DATA_PORT_OUTPUT;
	DATA_PORT = 0x00;
	// Set Control pins to outputs
	RS_PIN_TO_OUTPUT;
	RW_PIN_TO_OUTPUT;
        CLK_PIN_TO_OUTPUT;
	
	// Set initial conditions for control PIN'S
	RS_COMMAND;
	RW_WRITE;
	CLOCK_L;
	
	// What 15ms to make sure the display is powered up
	Delay15ms();
	// Run initialization sequence send 0x30 three times then enter 4 bit mode.
        PORTC |= 0x30;
        ClockPulseH();
        ClockPulseL();
        Delay5ms();
        // 2nd time
        ClockPulseH();
        ClockPulseL();
        // 3rd time
        ClockPulseH();
        ClockPulseL();
        // Setup 4 bit mode
        PORTC = PORTC & 0x0F;
        PORTC |= WORD_LENGTH_4;
        ClockPulseH();
        ClockPulseL();
       
        // Set the number of lines and character font
        // Select two line display with 5 x 8 characters
	while(BusyFlag());
            write(display_lines OR font_size);
	while(BusyFlag());
		DisplayOff();
	while(BusyFlag());
		ClearDisplay();
	while(BusyFlag());
		SetEntryPoint();
	while(BusyFlag());
		HomeCursor();
}

void write(unsigned char byte)
{
    int8_t high_nibble = byte & 0xF0;
    int8_t low_nibble = (byte << 4) & 0xF0;

    DATA_PORT = DATA_PORT & 0x0F;
    DATA_PORT |= high_nibble;
    ClockPulseH();
    ClockPulseL();

    DATA_PORT = DATA_PORT & 0x0F;
    DATA_PORT |= low_nibble;
    ClockPulseH();
    ClockPulseL();
}

void SetEntryPoint()
{
	// Set entry point Fixed display
	write(INCR_RIGHT);
}

void CursorOn(void)
{
	write(CURSOR_ON);
}

void CursorOff(void)
{
	write(CURSOR_OFF);
}

void HomeCursor(void)
{
	write(CURSOR_HOME);
}

void ShiftCursorLeft(void)
{
	write(SHIFT_CURSOR_LEFT);
}

void ShiftCursorRight(void)
{
	write(SHIFT_CURSOR_RIGHT);
}

void CursorBlinkOn(void)
{
	write(BLINK_ON);
}

void CursorBlinkOff(void)
{
	write(BLINK_OFF);
}

void DisplayOn(void)
{
	write(DISPLAY_ON OR CURSOR_ON);
}

void DisplayOff(void)
{
	write(DISPLAY_ON OR CURSOR_OFF);
}

void ClearDisplay(void)
{
	write(CLEAR_DISPLAY);
}

void ShiftDisplayLeft(void)
{
	write(SHIFT_DISPLAY_LEFT);
}

void ShiftDisplayRight(void)
{
	write(SHIFT_DISPLAY_RIGHT);
}

unsigned char BusyFlag()
{
    int8_t byte = 0;
    int8_t low_nibble = 0;

    // Make data port as an input
    TRIS_DATA_PORTC = DATA_PORT_INPUT;
	
    // Set R/W pin to read and enable it
    RW_READ;
    ClockPulseH();
    byte = PORTC & 0xF0;
    ClockPulseL();
    ClockPulseH();
    ClockPulseH();
    low_nibble = PORTC & 0xF0;
    ClockPulseL();
    RW_WRITE;
    TRIS_DATA_PORTC = DATA_PORT_OUTPUT;

    // Check Busy Flag
    if(PORTC  & BUSY_FLAG_MASK)
        return 1;
    else
	return 0;

}

void SetCharRamAddr(unsigned char addr)
{
    write(CHAR_RAM_ADDR OR addr);
}

void SetDisplayRamAddr(unsigned char addr)
{
    write(DISP_RAM_ADDRESS OR addr);
}

void WriteDisplay(unsigned char *t,unsigned char number_of_bytes)
{
	unsigned char c = 0;

	RS_CHAR_DISPLAY;

	Delay150us();

	for(c=0;c < number_of_bytes;c++)
	{
		if(t[c] == 0)
			break;

		write( t[c]);
                Delay150us();
        }

	RS_COMMAND;
}

unsigned char ReadDisplay()
{
    int8_t byte = 0;
    int8_t low_nibble = 0;

    // Make data port as an input
    TRIS_DATA_PORTC = DATA_PORT_INPUT;

    // Set R/W pin to read and enable it
    RW_READ;
    ClockPulseH();
    byte = PORTC & 0xF0;
    ClockPulseL();
    ClockPulseH();
    ClockPulseH();
    low_nibble = PORTC & 0xF0;
    ClockPulseL();
    RW_WRITE;
    TRIS_DATA_PORTC = DATA_PORT_OUTPUT;

    low_nibble = low_nibble >> 4;
    byte |= low_nibble;
    return byte;
}

void ClockPulseH()
{
	// Clock pulse > 450ns 
	CLOCK_H;
	__delay_us(150);
}

void ClockPulseL()
{
	// clock pulse > 450ns
	CLOCK_L;
	__delay_us(150);
}

void Delay15ms()
{
	__delay_ms(15);
}

void Delay5ms()
{
	__delay_ms(5);
}

void Delay150us()
{
	__delay_us(150);	
}
