// LCD_Driver.c Implementation of the HD44780 or simular LCD driver.
// The driver uses an 8 BIT DATA WORD. The DATA WORD is setup to use PORT D. To change it to another port
// Set DATA_PORT,TRIS_DATA_PORTD definition to new port in the header file.

// The contol pins are setup to use PORT E. To Change the port to another port change the following control pin
// definitions in the header file.
// #define RS_COMMAND 		RE0 = 0
// #define RS_CHAR_DISPLAY 	RE0 = 1
// #define RW_WRITE			RE1 = 0
// #define RW_READ 			RE1 = 1
// #define CLOCK_L 			RE2 = 0
// #define CLOCK_H 			RE2 = 1

// #define RS_PIN_TO_OUTPUT  TRISE0 = 0
// #define RW_PIN_TO_OUTPUT  TRISE1 = 0
// #define CLK_PIN_TO_OUTPUT TRISE2 = 0

// There are 3 delay function in the library these need to be ammened for the clock being used. The fuctions are
// setup to give the delays with a 48MHz clock. The delay calls are based on the number of instruction cycles.
// 1 instrction cycle = 4/clock frequency.

// The delay fuction which need to be changed are
// ClockPulseH()
// ClockPulseL()
// Delay15ms()
// Delay5ms()
// Delay150us()

// To use the libary first run the  InitDisplay(unsigned char display_lines,unsigned char font_size) function
// and set the number of display lines options are
// DISPLAY_LINES_1
// DISPLAY_LINES_2

// and the font type supported by the display. The options are
// FONT_5x8_DOT
// FONT_5x10_DOT

// Once you have initialized the display you can use any of the function avaliable in the library.
// Example to turn the display on and write a string to the display.

//	InitDisplay(DISLPAY_LINES_2,FONT_5x8_DOT);
		
//		while(BusyFlag());
//			DisplayOn();
	
//		sprintf((unsigned char *)t,"Wood & Douglas  ");
//		number_of_bytes = sizeof((char *)t);

//		while(BusyFlag());
//			WriteDisplay(&t,number_of_bytes);

// Use the while(BusyFlag()) to check the display can accept a new command.

#include <xc.h>
#include <pic16.h>
#include "lcddriver.h"

// PIC crystal frequency Hz or MHz
#define _XTAL_FREQ 3580000

void InitDisplay(unsigned char display_lines,unsigned char font_size)
{
	// Set Data word port to output
	TRIS_DATA_PORTD = DATA_PORT_OUTPUT;
	DATA_PORT = 0x00;
	// Set Control pins to outputs
	RS_PIN_TO_OUTPUT;
	RW_PIN_TO_OUTPUT;
        CLK_PIN_TO_OUTPUT;
	
	// Set initial conditions for control PIN'S
	RS_COMMAND;
	RW_WRITE;
	CLOCK_L;
	
	// What for display to inialize 15ms
	Delay15ms();

	// Initialization sequence.
	// Set 8bit operation and select two line display with 5 x 8
	DATA_PORT = WORD_LENGTH_8;
	ClockPulseH();
	ClockPulseL();
	Delay5ms();

	DATA_PORT = WORD_LENGTH_8;
	ClockPulseH();
	ClockPulseL();
	// Wait for instruction to excecute
	Delay150us();

	DATA_PORT = WORD_LENGTH_8;
	ClockPulseH();
	ClockPulseL();
	// Wait for instruction to excecute
	Delay150us();

	// Set the number of lines and character font
	DATA_PORT = WORD_LENGTH_8 OR display_lines OR font_size;
	ClockPulseH();
	ClockPulseL();

	while(BusyFlag());
		DisplayOff();
	while(BusyFlag());
		ClearDisplay();
	while(BusyFlag());
		SetEntryPoint();
	while(BusyFlag());
		HomeCursor();
}

void SetEntryPoint()
{
	// Set entry point Fixed display
	DATA_PORT = INCR_RIGHT;
	ClockPulseH();
	ClockPulseL();
}

void CursorOn(void)
{
	DATA_PORT = CURSOR_ON;
	ClockPulseH();
	ClockPulseL();
}

void CursorOff(void)
{
	DATA_PORT = CURSOR_OFF;
	ClockPulseH();
	ClockPulseL();
}

void HomeCursor(void)
{
	DATA_PORT = CURSOR_HOME;
	ClockPulseH();
	ClockPulseL();
}

void ShiftCursorLeft(void)
{
	DATA_PORT = SHIFT_CURSOR_LEFT;
	ClockPulseH();
	ClockPulseL();
}

void ShiftCursorRight(void)
{
	DATA_PORT = SHIFT_CURSOR_RIGHT;
	ClockPulseH();
	ClockPulseL();
}

void CursorBlinkOn(void)
{
	DATA_PORT = BLINK_ON;
	ClockPulseH();
	ClockPulseL();
}

void CursorBlinkOff(void)
{
	DATA_PORT = BLINK_OFF;
	ClockPulseH();
	ClockPulseL();
}

void DisplayOn(void)
{
	DATA_PORT = DISPLAY_ON OR CURSOR_ON;
	ClockPulseH();
	ClockPulseL();
}

void DisplayOff(void)
{
	DATA_PORT = DISPLAY_ON OR CURSOR_OFF;
	ClockPulseH();
	ClockPulseL();
}

void ClearDisplay(void)
{
	DATA_PORT = CLEAR_DISPLAY;
	ClockPulseH();
	ClockPulseL();
}

void ShiftDisplayLeft(void)
{
	DATA_PORT = SHIFT_DISPLAY_LEFT;
	ClockPulseH();
	ClockPulseL();
}

void ShiftDisplayRight(void)
{
	DATA_PORT = SHIFT_DISPLAY_RIGHT;
	ClockPulseH();
	ClockPulseL();
}

unsigned char BusyFlag()
{
	// Make data port as an input
	TRIS_DATA_PORTD = DATA_PORT_INPUT;
	
	// Set R/W pin to read and enable it
	RW_READ;
	ClockPulseH();

	// Wait for instruction to excecute
	Delay150us();
	
	// Check Busy Flag
	if(DATA_PORT & BUSY_FLAG_MASK)
	{
		ClockPulseL();
		RW_WRITE;
		TRIS_DATA_PORTD	= DATA_PORT_OUTPUT;
		return 1;
	}
	else
	{
		ClockPulseL();
		RW_WRITE;
		TRIS_DATA_PORTD = DATA_PORT_OUTPUT;
		return 0;
	}
}

void SetCharRamAddr(unsigned char addr)
{
	DATA_PORT = addr;
	ClockPulseH();
	ClockPulseL();
}

void WriteDisplay(unsigned char *t,unsigned char number_of_bytes)
{
	unsigned char c = 0;

	RS_CHAR_DISPLAY;

	Delay150us();

	for(c=0;c <= number_of_bytes;c++)
	{
		if(t[c] == 0)
			break;

		DATA_PORT = t[c];
		ClockPulseH();
		ClockPulseL();
		Delay150us();
	}

	RS_COMMAND;
}

void ReadDisplay(unsigned char *t)
{
	// Make data port as an input
	TRIS_DATA_PORTD = DATA_PORT_INPUT;
	
	RS_CHAR_DISPLAY;
	RW_READ;
	
	ClockPulseH();
	Delay150us();

	*t = PORTD;
	
	ClockPulseL();
	RW_WRITE;
	RS_COMMAND;
	// Make port output	
	TRIS_DATA_PORTD = DATA_PORT_OUTPUT;
}

void ClockPulseH()
{
	// Clock pulse > 450ns 
	CLOCK_H;
	__delay_us(500);
}

void ClockPulseL()
{
	// clock pulse > 450ns
	CLOCK_L;
	__delay_us(500);
}

void Delay15ms()
{
	__delay_ms(15);
}

void Delay5ms()
{
	__delay_ms(5);
}

void Delay150us()
{
	__delay_us(150);	
}
